/* linereader.c */
#include <stdlib.h>
#include <string.h>
#include "linereader.h"
#include "err.h" /* for CHK_NULL */

#define COMPRESS_WS 1
#define CONTINUATIONS 1
#define ACCEPT_NOCR 1

LR_input_t LR_init(LR_readfunc_t readfunc, void *where, char *str) {
  LR_input_t retval;
  retval.eof = 0;
  retval.cnt = strlen(str);
  retval.size = (retval.cnt < 256)? 256 : retval.cnt;
  retval.buf = malloc(retval.size); CHK_NULL(retval.buf);

  retval.readfunc = readfunc;
  retval.where = where;

  return retval;
}

/* compress spaces in a string, in place */
static char * strcompress(char *str) {
  char *src=str, *dst=str;
  while (*src!='\0') {
    int saw_space=0;
    while (*src==' '||*src=='\r'||*src=='\n'||*src=='\t') {
      saw_space=1; src++;
    }
    if (saw_space) *dst++=' ';
    *dst++=*src++;
  }
  *dst='\0';
  return str;
}

char * LR_readline(LR_input_t * in) {
  int n=0, sawCR=0, sawCRLF=0;
  int err;
  if (in->eof) return NULL;
 scan:
  for ( ; n < in->cnt; n++) {
    if ((CONTINUATIONS && sawCRLF && in->buf[n]!=' ' && in->buf[n]!='\t') ||
	((!CONTINUATIONS||n==1) && sawCR  && in->buf[n]=='\n')
#ifdef ACCEPT_NOCR /* but no continuations... */
	|| (!sawCR && in->buf[n]=='\n')
#endif
	)  { /* done */
      char tmp[n+1];
      /* move past \n if necessary */
      if ((!CONTINUATIONS||n==1) && sawCR  && in->buf[n]=='\n') n++;
#ifdef ACCEPT_NOCR
      if (!sawCR && in->buf[n]=='\n') n++;
#endif
      /* adjust buffer count. */
      in->cnt-=n;
      /* move line to end of buffer */
      memcpy(tmp, in->buf, n);
      memcpy(in->buf, in->buf+n, in->cnt);
      memcpy(in->buf+in->cnt, tmp, n);
      /* null-terminate */
      if (n>1 && in->buf[in->cnt+n-2]=='\r')
	in->buf[in->cnt+n-2]='\0';
      else
	in->buf[in->cnt+n-1]='\0';
      if (COMPRESS_WS) strcompress(in->buf + in->cnt);
      return in->buf + in->cnt;
    }
    sawCRLF=sawCR && (in->buf[n]=='\n');
    sawCR=(in->buf[n]=='\r');
  }
  /* got to end of in->buffer without finding NL */
  if (in->cnt==in->size) { /* expand in->buffer: one huge line! */
    in->size*=2;
    in->buf = realloc(in->buf, in->size); CHK_NULL(in->buf);
  }
  err=in->readfunc(in->where, in->buf + in->cnt, in->size - in->cnt);
  if (err>0) { in->cnt+=err; goto scan; }
  // okay, eof.
  in->eof = 1;
  if (n==0) return NULL;
  in->buf[n+1]='\0'; /* guaranteed there's room because n==cnt<size */
  if (COMPRESS_WS) strcompress(in->buf);
  return in->buf;
}

void *LR_rawread(LR_input_t *in, int * length) {
  if (in->eof) { *length = in->cnt; in->cnt=0; return in->buf; }
  if (in->size < *length) {
    in->size = *length;
    in->buf = realloc(in->buf, in->size); CHK_NULL(in->buf);
  }
  while (in->cnt < *length) {
    int err=in->readfunc(in->where, in->buf + in->cnt, *length - in->cnt);
    if (err<=0) { in->eof=1; *length = in->cnt; in->cnt=0; return in->buf; }
    else in->cnt+=err;
  }
  in->cnt -= *length;
  if (in->cnt > 0) { /* move raw data to end of buffer */
    char tmp[*length];
    memcpy(tmp, in->buf, *length);
    memcpy(in->buf, in->buf+*length, in->cnt);
    memcpy(in->buf+in->cnt, tmp, *length);
    return in->buf+in->cnt;
  } else return in->buf;
}

void LR_free(LR_input_t * in) {
  if (in->buf != NULL) { free(in->buf); in->buf=NULL; }
}

